"""Unit tests for the Lytspel converter."""

# pylint: disable=invalid-name, line-too-long, missing-docstring, redefined-outer-name, protected-access
# Note: invalid-name disabled since pylint limits function names to 30 chars.

from collections import Counter
from pytest import fixture

from . import Converter
from .conv import WORD_RE
from .dict import ConvState, Dictionary

# Fixtures

@fixture(scope='session')
def dct():
    return Dictionary()

@fixture(scope='session')
def conv():
    return Converter()

@fixture
def is_word():
    """'is_word' was formerly a function.

    But it has been replaced with a direct regex call for efficiency."""
    return WORD_RE.match

# Tests

def test_lookup_simple(dct):
    assert dct.lookup('blackbird') == 'blakburd'
    assert dct.lookup('earthward') == 'urthwerd'
    assert dct.lookup('shadows') == 'shadoas'
    assert dct.lookup('sharp') == 'sharp'
    assert dct.lookup('technology') == 'tec’noleji'
    assert dct.lookup('tahiti') == 'ta’heeti'   # capitalized in dict
    assert dct.lookup('notaword') is None

def test_lookup_case_restored(dct):
    """Case should be restored."""
    assert dct.lookup('Score') == 'Scoar'
    assert dct.lookup('NETWORK') == 'NETWURK'
    assert dct.lookup('POLITICAL') == 'PO’LITICL'
    assert dct.lookup('JavaScript') == 'JaavaScript'
    assert dct.lookup('McCain') == 'McCain'
    assert dct.lookup('MeV') == 'MeV'
    assert dct.lookup('ODed') == 'ODed'
    assert dct.lookup("O'Donnell") == 'O’Donel'
    assert dct.lookup('PCs') == 'PCs'
    assert dct.lookup('PowerPC') == 'PowerPC'

def test_lookup_apostrophes(dct):
    """Both normal and typographic apostrophes should be accepted."""
    assert dct.lookup("we'd") == 'wi’d'
    assert dct.lookup('we’d') == 'wi’d'
    assert dct.lookup("it'll") == 'it’l'
    assert dct.lookup('it’ll') == 'it’l'
    assert dct.lookup("I'm").lower() == 'y’m'
    assert dct.lookup('I’m').lower() == 'y’m'
    assert dct.lookup("O'Connell") == 'O’Conel'
    assert dct.lookup('O’Connell') == 'O’Conel'

def test_lookup_genitive_s(dct):
    """'s (genitive or contraction) is handled correctly (normal and typographic forms)."""
    assert dct.lookup("He's") == 'Hi’s'
    assert dct.lookup("boyfriend's") == 'boyfrend’s'
    assert dct.lookup('He’s') == 'Hi’s'
    assert dct.lookup('boyfriend’s') == 'boyfrend’s'
    # Upper case words too
    assert dct.lookup("HE'S") == 'HI’S'
    assert dct.lookup("HE's") == 'HI’s'

def test_lookup_contractions(dct):
    assert dct.lookup("'s") == '’s'
    assert dct.lookup("’s") == '’s'
    assert dct.lookup("'d") == '’d'
    assert dct.lookup("’d") == '’d'
    assert dct.lookup("'ll") == '’l'
    assert dct.lookup("'re") == '’r'
    assert dct.lookup("'ve") == '’v'
    assert dct.lookup("’ve") == '’v'

def test_lookup_nonstandard_contractions(dct):
    assert dct.lookup('Peter’d') == 'Peeter’d'
    assert dct.lookup('PETER’D') == 'PEETER’D'
    assert dct.lookup("what'll") == 'wot’l'
    assert dct.lookup("What'll") == 'Wot’l'
    assert dct.lookup('What’Re') == 'Wot’R'
    assert dct.lookup("where've") == 'wair’v'
    assert dct.lookup("WHERE'VE") == 'WAIR’V'

def test_lookup_diacritics(dct):
    assert dct.lookup('café') == 'ca’fay'
    assert dct.lookup('continuüm') == 'con’tiniuam'
    assert dct.lookup('doppelgänger') == 'dopelganger'
    assert dct.lookup('élite') == 'i’leet'
    assert dct.lookup('Élite') == 'I’leet'
    assert dct.lookup('épée') == 'ai’pay'
    assert dct.lookup('ÉPÉE') == 'AI’PAY'
    assert dct.lookup('hôtel') == 'hoa’tel'
    assert dct.lookup('Hôtel') == 'Hoa’tel'
    assert dct.lookup('mañana') == 'maa’nyaanaa'
    assert dct.lookup('naïve') == 'naa’eev'
    assert dct.lookup('œuvre') == 'uuvra'
    assert dct.lookup('Œuvre') == 'Uuvra'
    assert dct.lookup('tōfu') == 'toafu'

def test_lookup_redirects(dct):
    """Test that redirects are resolved correctly."""
    assert dct.lookup('Anglicised') == 'Anglicysd'
    assert dct.lookup('barques') == 'barks'
    assert dct.lookup('calking') == 'cauking'
    assert dct.lookup('castors') == 'casters'
    assert dct.lookup('Centre') == 'Senter'
    assert dct.lookup('aesthetician') == 'esthi’tition'
    assert dct.lookup('ESOP') == 'EESSOP'
    assert dct.lookup('Ier') == 'Yr'
    assert dct.lookup('OKs') == 'Oa’cays'
    assert dct.lookup('paralyses') == 'parrelyses'
    assert dct.lookup('prise') == 'prys'
    assert dct.lookup('reanalyses') == 'ri’anelyses'

def test_lookup_nlp_needed_pos(dct):
    """Test that lookup signals its need for a POS tag if one is necessary but not given."""
    assert dct.lookup('increase') is ConvState.NLP_NEEDED
    assert dct.lookup('misuse') is ConvState.NLP_NEEDED

def test_lookup_pos_tagged(dct):
    """Test that POS-tagged words are looked up correctly."""
    assert dct.lookup('increase', 'NOUN') == 'increess'
    assert dct.lookup('increase', 'VERB') == 'in’creess'
    assert dct.lookup('misuse', 'NOUN') == 'mis’iuss'
    assert dct.lookup('misuse', 'VERB') == 'mis’ius'

def test_lookup_hyphenated_prefix(dct):
    assert dct.lookup('re-') == 'ri-'
    assert dct.lookup('Re-') == 'Ri-'
    assert dct.lookup('RE-') == 'RI-'
    assert dct.lookup('de-') == 'di-'
    assert dct.lookup('meta-') == 'mete-'
    assert dct.lookup('paleo-') == 'palio-'
    assert dct.lookup('Paleo-') == 'Palio-'
    assert dct.lookup('PALEO-') == 'PALIO-'

def test_lookup_us(dct):
    """A few capitalized abbreviation such as 'US' should remain unchanged."""
    assert dct.lookup('us') == 'uss'
    assert dct.lookup('Us') == 'Uss'
    assert dct.lookup('US') == 'US'
    assert dct.lookup('US') == 'US'
    assert dct.lookup("US's") == 'US’s'
    assert dct.lookup('US’s') == 'US’s'

def test_lookup_in(dct):
    """Accept -in (usually followed by a contraction marker) as an alternative to -ing."""
    assert dct.lookup('friggin') == 'frigin'
    assert dct.lookup('frigging') == 'friging'
    assert dct.lookup('singin') == 'singin'
    assert dct.lookup('Singin') == 'Singin'
    assert dct.lookup('SINGIN') == 'SINGIN'
    assert dct.lookup('singing') == 'singing'
    # Unless the respelled -ing word does NOT end in -ing
    assert dct.lookup('yealin') is None
    assert dct.lookup('yealing') == 'yeelin'


def test_tokenize_simple(conv):
    assert '|'.join(conv.tokenize_text('This is a sentence.')) == 'This| |is| |a| |sentence|.'
    assert '|'.join(conv.tokenize_text('A sentence without final punctuation')) ==\
            'A| |sentence| |without| |final| |punctuation'
    assert '|'.join(conv.tokenize_text('Sentence with-some inner, punctuation;this should - not! — cause problems?Let    us\thope so!')) ==\
            'Sentence| |with|-|some| |inner|, |punctuation|;|this| |should| - |not|! — |cause| |problems|?|Let|    |us|\t|hope| |so|!'
    assert '|'.join(conv.tokenize_text('Some words in "double" and \'half quotation\' marks.')) ==\
            'Some| |words| |in| "|double|" |and| \'|half| |quotation|\' |marks|.'
    assert '|'.join(conv.tokenize_text('Some words in typographic “double” and ‘half quotation’ marks.')) ==\
            'Some| |words| |in| |typographic| “|double|” |and| ‘|half| |quotation|’ |marks|.'
    assert '|'.join(conv.tokenize_text('This is a sentence.')) == 'This| |is| |a| |sentence|.'
    assert '|'.join(conv.tokenize_text(':::sentence with leading punctuation:::')) ==\
            ':::|sentence| |with| |leading| |punctuation|:::'

def test_tokenize_contractions(conv):
    assert '|'.join(conv.tokenize_text("Let's hope contractions are handled correctly wheresoe'er they'll occur, don't you think so, O'Connell?")) ==\
        "Let's| |hope| |contractions| |are| |handled| |correctly| |wheresoe'er| |they'll| |occur|, |don't| |you| |think| |so|, |O'Connell|?"
    assert '|'.join(conv.tokenize_text('Let’s hope contractions are handled correctly wheresoe’er they’ll occur, don’t you think so, O’Connell?')) ==\
        'Let’s| |hope| |contractions| |are| |handled| |correctly| |wheresoe’er| |they’ll| |occur|, |don’t| |you| |think| |so|, |O’Connell|?'
    assert '|'.join(conv.tokenize_text("He's happy to see my boyfriend's sister.")) ==\
        "He's| |happy| |to| |see| |my| |boyfriend's| |sister|."
    assert '|'.join(conv.tokenize_text('He’s happy to see my boyfriend’s sister.')) ==\
        'He’s| |happy| |to| |see| |my| |boyfriend’s| |sister|.'

def test_tokenize_diacritics(conv):
    assert '|'.join(conv.tokenize_text('Mañana me and my naïve doppelgänger will eat tōfu in the café of an élite hôtel.')) ==\
      'Mañana| |me| |and| |my| |naïve| |doppelgänger| |will| |eat| |tōfu| |in| |the| |café| |of| |an| |élite| |hôtel|.'
    assert '|'.join(conv.tokenize_text('MAÑANA me and my naïve doppelGÄNGER will eat tōfu in the CAFÉ of an Élite Hôtel.')) ==\
      'MAÑANA| |me| |and| |my| |naïve| |doppelGÄNGER| |will| |eat| |tōfu| |in| |the| |CAFÉ| |of| |an| |Élite| |Hôtel|.'

def test_tokenize_dates_and_numbers(conv):
    assert '|'.join(conv.tokenize_text('On 11 Feb. 2019 I wrote a test.')) ==\
            'On| 11 |Feb|. 2019 |I| |wrote| |a| |test|.'
    assert '|'.join(conv.tokenize_text('33 divided by 2 is 16.5.')) ==\
            '33 |divided| |by| 2 |is| 16.5.'


def test_convert_para_simple(conv):
    assert conv.convert_para('This is a sentence.') == 'Dhiss is a sentenss.'
    assert conv.convert_para('Some words in "double" and \'half quotation\' marks.') ==\
        'Sum wurds in "dubl" and \'haf quoa’taition\' marks.'
    assert conv.convert_para('Some words in typographic “double” and ‘half quotation’ marks.') ==\
        'Sum wurds in typo’grafic “dubl” and ‘haf quoa’taition’ marks.'

def test_convert_para_spacy_quotes(conv):
    """Test that quote marks are also handled correctly if spaCy is invoked."""
    assert conv.convert_para('Who could object to some words in "double" and \'half quotation\' marks?') ==\
        'Hu cood ob’ject tu sum wurds in "dubl" and \'haf quoa’taition\' marks?'
    assert conv.convert_para('Who could object to some words in typographic “double” and ‘half quotation’ marks?') ==\
        'Hu cood ob’ject tu sum wurds in typo’grafic “dubl” and ‘haf quoa’taition’ marks?'
    assert conv.convert_para('Who could object to some words in typographic ‘‘fake double quotation’’ marks?') ==\
        'Hu cood ob’ject tu sum wurds in typo’grafic ‘‘faik dubl quoa’taition’’ marks?'

def test_convert_para_contractions(conv):
    assert conv.convert_para("Let's hope contractions are handled correctly wheresoe'er they'll occur, don't you think so, O'Connell?") ==\
        'Let’s hoap con’tractions ar handld ke’rectli wairso’air dhay’l o’cur, doan’t iu think so, O’Conel?'
    assert conv.convert_para('Let’s hope contractions are handled correctly wheresoe’er they’ll occur, don’t you think so, O’Connell?') ==\
        'Let’s hoap con’tractions ar handld ke’rectli wairso’air dhay’l o’cur, doan’t iu think so, O’Conel?'
    assert conv.convert_para("He's happy to see my boyfriend's sister.") ==\
        'Hi’s hapi tu see my boyfrend’s sister.'
    assert conv.convert_para('He’s happy to see my boyfriend’s sister.') ==\
        'Hi’s hapi tu see my boyfrend’s sister.'

def test_convert_para_nonstandard_contractions(conv):
    """Ensure that nonstandard contractions are accepted as well."""
    assert conv.convert_para("Sue'll do it 'cause she can.") ==\
        "Su’l du it 'caus shi can."
    assert conv.convert_para('Sue’ll do it ’cause she can.') ==\
        'Su’l du it ’caus shi can.'
    assert conv.convert_para('Peter’d know what to do.') ==\
        'Peeter’d noa wot tu du.'
    assert conv.convert_para("What're they going to do?") ==\
        'Wot’r dhay going tu du?'
    assert conv.convert_para('Where’ve they put it?') ==\
        'Wair’v dhay poot it?'

def test_convert_para_nonstandard_contractions_spacy(conv):
    """Likewise with spaCy."""
    assert conv.convert_para("I object: Sue'll do it 'cause she can.") ==\
        'Y ob’ject: Su’l du it ’caus shi can.'
    assert conv.convert_para('I object: Sue’ll do it ’cause she can.') ==\
        'Y ob’ject: Su’l du it ’caus shi can.'
    assert conv.convert_para('I estimate: Peter’d know what to do.') ==\
        'Y estimait: Peeter’d noa wot tu du.'
    assert conv.convert_para("I object: What're they going to do?") ==\
        'Y ob’ject: Wot’r dhay going tu du?'

def test_convert_para_pos_tagged(conv):
    assert conv.convert_para('I did not object to the object.') == 'Y did not ob’ject tu dhe object.'
    assert conv.convert_para('They were too close to the door to close it.') ==\
        'Dhay wur tuu cloass tu dhe doar tu cloas it.'
    assert conv.convert_para('Before I mow the lawn let me place this grain in the mow.') ==\
        'Bi’foar y mo dhe laun let mi plaiss dhiss grain in dhe mow.'
    assert conv.convert_para('He thought it was time to present the present.') ==\
        'Hi thaut it wos tym tu pri’sent dhe present.'
    assert conv.convert_para('I met an august man last August.') ==\
        'Y met an au’gust man last August.'
    assert conv.convert_para('To help with planting, the farmer taught his sow to sow.') ==\
        'Tu help widh planting, dhe farmer taut his sow tu so.'
    assert conv.convert_para('The weather was beginning to affect his affect.') ==\
        'Dhe wedher wos bi’gining tu a’fect his afect.'
    assert conv.convert_para('We must polish the Polish furniture.') ==\
        'Wi must polish dhe Poalish furnicher.'
    assert conv.convert_para('The dump was so full that it had to refuse more refuse.') ==\
        'Dhe dump wos so fool dhat it had tu ri’fius moar refiuss.'
    assert conv.convert_para('I had to subject the subject to a series of tests.') ==\
        'Y had tu sub’ject dhe subject tu a seerees ov tests.'
    assert conv.convert_para('Don’t desert me here in the desert!') ==\
        'Doan’t di’surt mi heer in dhe desert!'
    assert conv.convert_para('The outright prohibition has caused smoking to be banned outright.') ==\
            'Dhe outryt prohi’bition has causd smoaking tu bee band out’ryt.'

def test_convert_para_nt_contractions(conv):
    assert conv.convert_para("Don't you think I won't do it, because I will!") ==\
            'Doan’t iu think y woan’t du it, bi’caus y wil!'
    assert conv.convert_para('Don’t you think I won’t do it, because I will!') ==\
            'Doan’t iu think y woan’t du it, bi’caus y wil!'
    assert conv.convert_para("You mustn't believe that they can't do such a thing.") ==\
            'Iu musn’t bi’leev dhat dhay can’t du such a thing.'
    assert conv.convert_para('You mustn’t believe that they can’t do such a thing.') ==\
            'Iu musn’t bi’leev dhat dhay can’t du such a thing.'

def test_convert_para_initial_nt(conv):
    """Leading "n't", though certainly irregular, shouldn't cause problems.

    Rather, it should be treated as an unknown token and returned as is.
    """
    assert conv.convert_para("N't a good way to open a sentence.") ==\
            "N't a good way tu oapen a sentenss."
    assert conv.convert_para('N’t a good way to open a sentence.') ==\
            'N’t a good way tu oapen a sentenss.'

def test_convert_para_diacritics(conv):
    assert conv.convert_para('Mañana me and my naïve doppelgänger will eat tōfu in the café of an élite hôtel.') ==\
      'Maa’nyaanaa mi and my naa’eev dopelganger wil eet toafu in dhe ca’fay ov an i’leet hoa’tel.'


def test_convert_i_case(conv):
    """Test case of converted "I" and contractions such as "I'd", "I'll".

    They should be capitalized at the start, but not in the middle of sentences.
    """
    assert conv.convert_para("I am capitalized at the start of sentences but I'm lower-case in the middle. I am still capitalized at the start.") ==\
            'Y am capitelysd at dhe start ov sentensses but y’m loer-caiss in dhe midl. Y am stil capitelysd at dhe start.'
    assert conv.convert_para('I’d be capitalized at the start of sentences but I’ll be lower-case in the “middle”! I’d still be capitalized at the start.') ==\
            'Y’d bee capitelysd at dhe start ov sentensses but y’l bee loer-caiss in dhe “midl”! Y’d stil bee capitelysd at dhe start.'
    assert conv.convert_para("I've seen that that works but I've not seen whether this works? I've seen it now.") ==\
            'Y’v seen dhat dhat wurks but y’v not seen wedher dhiss wurks? Y’v seen it now.'

def test_convert_i_case_quotes(conv):
    """Test "I" case conversion after quote marks."""
    assert conv.convert_para('Opening quote marks don\'t hurt, I guess. "I am still capitalized."') ==\
            'Oapening quoat marks doan’t hurt, y gess. "Y am stil capitelysd."'
    assert conv.convert_para('Opening quote marks don’t hurt, I guess. “I am still capitalized.”') ==\
            'Oapening quoat marks doan’t hurt, y gess. “Y am stil capitelysd.”'
    assert conv.convert_para('"Closing trailing quote marks don\'t hurt, I hope?" I\'m still capitalized.') ==\
            '"Cloasing trailing quoat marks doan’t hurt, y hoap?" Y’m stil capitelysd.'
    assert conv.convert_para('“Closing trailing quote marks don’t hurt, I hope!” I’m still capitalized.') ==\
            '“Cloasing trailing quoat marks doan’t hurt, y hoap!” Y’m stil capitelysd.'
    assert conv.convert_para("Closing leading quote marks don't hurt, I 'hope'. I am still capitalized.") ==\
            "Cloasing leeding quoat marks doan’t hurt, y 'hoap'. Y am stil capitelysd."
    assert conv.convert_para('Closing leading quote marks don’t hurt, I ‘hope’! I am still capitalized.') ==\
            'Cloasing leeding quoat marks doan’t hurt, y ‘hoap’! Y am stil capitelysd.'

def test_convert_i_case_colon(conv):
    """Test "I" case conversion after a colon.

    A quoted phrase after a colon is assumed to start a new sentence (capitalized).
    In all other cases, the preceding sentence is assumed to continue (lower-case).
    """
    assert conv.convert_para('After a colon: I should not be capitalized.') ==\
            'After a coalen: y shood not bee capitelysd.'
    assert conv.convert_para('I said: "I hope I\'ll be capitalized now."') ==\
            'Y sed: "Y hoap y’l bee capitelysd now."'
    assert conv.convert_para('I said: “I hope I’ll be capitalized now.”') ==\
            'Y sed: “Y hoap y’l bee capitelysd now.”'
    assert conv.convert_para('I said: ‘I hope I’ll be capitalized now.’') ==\
            'Y sed: ‘Y hoap y’l bee capitelysd now.’'

def test_convert_i_case_dates_and_numbers(conv):
    """Test "I" case conversion after dates and numbers."""
    assert conv.convert_para('On 11 Feb. 2019 I wrote a test.') ==\
            'On 11 Feb. 2019 y roat a test.'
    assert conv.convert_para('Even after a fraction such as 3.5 I should still be lower-case.') ==\
            'Eeven after a fraction such as 3.5 y shood stil bee loer-caiss.'

def test_convert_i_case_spacy(conv):
    """Test that "I" case conversion also works if spaCy is invoked."""
    assert conv.convert_para("I estimate that I am capitalized at the start of sentences but that I'm lower-case in the middle. I am still capitalized at the start.") ==\
            'Y estimait dhat y am capitelysd at dhe start ov sentensses but dhat y’m loer-caiss in dhe midl. Y am stil capitelysd at dhe start.'
    assert conv.convert_para('Opening quote marks don’t hurt, I estimate. “I am still capitalized.”') ==\
            'Oapening quoat marks doan’t hurt, y estimait. “Y am stil capitelysd.”'
    assert conv.convert_para('"Closing trailing quote marks don\'t hurt, I estimate?" I\'m still capitalized.') ==\
            '"Cloasing trailing quoat marks doan’t hurt, y estimait?" Y’m stil capitelysd.'
    assert conv.convert_para('Closing leading quote marks don’t hurt, ‘I’d estimate’! I am still capitalized.') ==\
            'Cloasing leeding quoat marks doan’t hurt, ‘y’d estimait’! Y am stil capitelysd.'
    assert conv.convert_para('After a colon: I should not be capitalized, I’d estimate.') ==\
            'After a coalen: y shood not bee capitelysd, y’d estimait.'
    assert conv.convert_para('I said: “I estimate I’ll be capitalized now.”') ==\
            'Y sed: “Y estimait y’l bee capitelysd now.”'
    assert conv.convert_para('I estimate that around 11 Feb. 2019 I wrote a test.') ==\
            'Y estimait dhat e’round 11 Feb. 2019 y roat a test.'
    assert conv.convert_para('Even after a fraction such as 3.5 I should still be lower-case, I estimate.') ==\
            'Eeven after a fraction such as 3.5 y shood stil bee loer-caiss, y estimait.'

def test_convert_i_vs_roman_i(conv):
    """Test that the pronoun 'I' and the roman number 'I' are distinguished."""
    assert conv.convert_para('I survived World War I, but many others didn’t.') ==\
            'Y ser’vyvd Wurld Wor I, but meni udhers didn’t.'


def test_convert_hyphenated_prefixes(conv):
    assert conv.convert_para('I re-sent the file to you last night.') ==\
        'Y ri-sent dhe fyl tu iu last nyt.'
    # Test that it works with spaCy too
    assert conv.convert_para('I estimate that I have to re-press the shirt.') ==\
        'Y estimait dhat y hav tu ri-press dhe shurt.'
    # Also if capitalized
    assert conv.convert_para('The Re-Creation of a Lost Artwork') ==\
        'Dhe Ri-Cri’aition ov a Lost Artwurk'
    # Nothing should happen in other cases
    assert conv.convert_para('I talked with her re the case.') ==\
        'Y taukd widh hur ree dhe caiss.'
    assert conv.convert_para('Re: Your mail') ==\
        'Ree: Iur mail'
    # Also test some other prefix
    assert conv.convert_para('She de-emphasized our differences.') ==\
            'Shi di-emfecysd our diferensses.'
    assert conv.convert_para('The Paleo-Americans arrived in the Americas a long, long time ago.') ==\
        'Dhe Palio-A’merricans e’ryvd in dhe A’merricas a long, long tym a’go.'

def test_convert_other_hyphens(conv):
    """Test that hyphens between words don't cause problems."""
    # Hyphens between independent words
    assert conv.convert_para('We need a state-of-the-art solution.') ==\
        'Wi need a stait-ov-dhe-art so’luution.'
    # Not a prefix, since there is whitespace after the hyphen
    assert conv.convert_para('I’m contacting you re- you might have guessed it -your mail.') ==\
        'Y’m contacting iu ree- iu myt hav gessd it -iur mail.'
    # Likewise if another hyphen follows
    assert conv.convert_para('I’m contacting you re--you might have guessed it--your mail.') ==\
        'Y’m contacting iu ree--iu myt hav gessd it--iur mail.'

def test_convert_urls(conv):
    """Test that URLs are not converted."""
    assert conv.convert_para('This is a link: www.lytspel.org/overview, hence it should not be converted.') ==\
        'Dhiss is a link: www.lytspel.org/overview, henss it shood not bee con’vurted.'
    assert conv.convert_para('This is a link: https://www.lytspel.org/idea which is not converted.') ==\
        'Dhiss is a link: https://www.lytspel.org/idea wich is not con’vurted.'
    assert conv.convert_para('This is a link: <https://www.lytspel.org/limitations>, hence it should not be converted.') ==\
        'Dhiss is a link: <https://www.lytspel.org/limitations>, henss it shood not bee con’vurted.'
    assert conv.convert_para('See the subsequent link (https://www.lytspel.org/rules) which is not converted.') ==\
        'See dhe subsiquent link (https://www.lytspel.org/rules) wich is not con’vurted.'
    assert conv.convert_para('See the subsequent link [https://www.lytspel.org/overview], which is not converted.') ==\
        'See dhe subsiquent link [https://www.lytspel.org/overview], wich is not con’vurted.'

def test_convert_urls_spacy(conv):
    """Ditto with spaCy."""
    assert conv.convert_para('This, I object, is a link: www.lytspel.org/overview, hence it should not be converted.') ==\
        'Dhiss, y ob’ject, is a link: www.lytspel.org/overview, henss it shood not bee con’vurted.'
    assert conv.convert_para('This, I object, is a link: https://www.lytspel.org/idea which is not converted.') ==\
        'Dhiss, y ob’ject, is a link: https://www.lytspel.org/idea wich is not con’vurted.'
    assert conv.convert_para('This, I object, is a link: <https://www.lytspel.org/limitations>, hence it should not be converted.') ==\
        'Dhiss, y ob’ject, is a link: <https://www.lytspel.org/limitations>, henss it shood not bee con’vurted.'
    assert conv.convert_para('I object: See the subsequent link (https://www.lytspel.org/rules) which is not converted.') ==\
        'Y ob’ject: See dhe subsiquent link (https://www.lytspel.org/rules) wich is not con’vurted.'
    assert conv.convert_para('I object: See the subsequent link [https://www.lytspel.org/overview], which is not converted.') ==\
        'Y ob’ject: See dhe subsiquent link [https://www.lytspel.org/overview], wich is not con’vurted.'

def test_convert_emails(conv):
    """Test that email addressed are not converted."""
    assert conv.convert_para('This is an email: christian@example.org, hence it should not be converted.') ==\
        'Dhiss is an eemail: christian@example.org, henss it shood not bee con’vurted.'
    assert conv.convert_para('This is an email: mailto:christian@example.org which is not converted.') ==\
        'Dhiss is an eemail: mailto:christian@example.org wich is not con’vurted.'
    assert conv.convert_para('This is an email: <christian@example.org>, hence it should not be converted.') ==\
        'Dhiss is an eemail: <christian@example.org>, henss it shood not bee con’vurted.'
    assert conv.convert_para('See the subsequent email (christian@example.org) which is not converted.') ==\
        'See dhe subsiquent eemail (christian@example.org) wich is not con’vurted.'

def test_convert_emails_spacy(conv):
    """Ditto with spaCy."""
    assert conv.convert_para('This, I object, is an email: christian@example.org, hence it should not be converted.') ==\
        'Dhiss, y ob’ject, is an eemail: christian@example.org, henss it shood not bee con’vurted.'
    assert conv.convert_para('This, I object, is an email: mailto:christian@example.org which is not converted.') ==\
        'Dhiss, y ob’ject, is an eemail: mailto:christian@example.org wich is not con’vurted.'
    assert conv.convert_para('This, I object, is an email: <christian@example.org>, hence it should not be converted.') ==\
        'Dhiss, y ob’ject, is an eemail: <christian@example.org>, henss it shood not bee con’vurted.'
    assert conv.convert_para('I object: See the subsequent email (christian@example.org) which is not converted.') ==\
        'Y ob’ject: See dhe subsiquent eemail (christian@example.org) wich is not con’vurted.'

def test_is_word_simple(is_word):
    assert is_word('a')
    assert is_word('This')
    assert is_word('sentence')
    assert is_word("Let's")
    assert not is_word('')
    assert not is_word(' ')
    assert not is_word('.')
    assert not is_word('; ')
    assert not is_word("'")
    assert not is_word('’')
    assert not is_word('“')

def test_is_word_apostrophe(is_word):
    """Words starting with an apostrophe are recognized as such.

    SpaCy's tokenizer produces such words.
    """
    assert is_word("'d")
    assert is_word("'re")
    assert is_word('’d"')
    assert is_word('’re')

def test_is_word_diacritics(is_word):
    assert is_word('œuvre')
    assert is_word('élite')
    assert is_word('épée')
    assert is_word('doppelgänger')
    assert is_word('Œuvre')
    assert is_word('Élite')
    assert is_word('ÉPÉE')


def test_unify_case_differences_basic(conv):
    """Test that the counter works as it should."""
    conv._unknown_counter = Counter(hulla=3, Hulla=1, HULLA=1)
    conv._unify_case_differences()
    assert conv._unknown_counter == Counter(hulla=5)

def test_unify_case_differences_empty(conv):
    """Empty counter should not change."""
    conv._unknown_counter = Counter()
    conv._unify_case_differences()
    assert conv._unknown_counter == Counter()

def test_unify_case_differences_no_common_lower_case(conv):
    """Counter should not change if there is no common lower-case version."""
    conv._unknown_counter = Counter(Hulla=1, HULLA=1)
    conv._unify_case_differences()
    assert conv._unknown_counter == Counter(Hulla=1, HULLA=1)

def test_unify_case_differences_different_entries(conv):
    """Counter should not change if all entries differ even ignoring case."""
    conv._unknown_counter = Counter(alpha=1, Beta=2, GAMMA=3, delta=4)
    conv._unify_case_differences()
    assert conv._unknown_counter == Counter(alpha=1, Beta=2, GAMMA=3, delta=4)



"""Defines the dictionary mapping tradspell to Lytspel."""

import csv
from enum import Enum
from functools import lru_cache
from io import TextIOWrapper
import re
from typing import List, Tuple, Union
from warnings import warn

from pkg_resources import resource_stream

from .util import get_elem


# Non-ASCII letters which we know how to convert to ASCII (lower-case only).
HANDLED_NON_ASCII_LETTERS = 'áàăâǎåäãąāảæćĉčċçďđéèĕêěëẽęēẻǵğĝǧġḡĥȟḧħíìĭîǐïĩįīĵǰḱǩĺľłḹḿńǹňñ'\
                            'óòŏôǒöõøǫōỏœṕŕřśŝšşßťẗúùŭûǔůüũųūủṽẃẁẘẅẍýỳŷẙÿỹȳỷźẑž'

# Matches a non-ASCII letter which we know how to convert.
# Note: precompiling all regexes, while not as readable, actually speeds up the conversion
# process quite a bit.
NON_ASCII_LETTERS_RE = re.compile('[{}]'.format(HANDLED_NON_ASCII_LETTERS), re.IGNORECASE)

# Matches a capitalized letter followed by a lower-case one (optionally with something else
# between them).
CAPTALIZED_RE = re.compile('[A-Z].*[a-z]')

# Matches a capitalized letter occurring after something else.
INNER_CAP_RE = re.compile('.[A-Z]')

# Matches all-caps words.
ALL_CAPS_RE = re.compile("[-'A-Z]+$")

# Matches an upper-case letter.
UPPER_CASE_RE = re.compile('[A-Z]')

# Matches a single-letter contraction such as "'d".
SINGLE_LETTER_CONTRACTION_RE = re.compile("'.$")

# Matches 'I' and its contractions.
I_AND_CONTRACTIONS_RE = re.compile("I$|I['’]")

# Matches words ending in "'s" (genitive) or generic contractions such as "'ve".
GENERIC_CONTRACTION_RE = re.compile(".['’](s|d|ll|re|ve)$", re.IGNORECASE)

# All-caps abbreviations that should be kept unchanged
CAP_ABBREVS = frozenset(('BIOS', 'DE', 'ETA', 'GA', 'HI', 'LA', 'MA', 'ME', 'MI', 'OH', 'PA', 'US',
                         'WA'))

POS_FALLBACKS = {'aj': 'n', 'av': 'aj', 'n': 'aj', 'prp': 'aj', 'v': 'n'}

ConvState = Enum('ConvState', 'LOOKS_FOREIGN NLP_NEEDED')  # pylint: disable=invalid-name


class Dictionary:
    """A dictionary mapping tradspell to Lytspel.

    Instances of this class are threadsafe after initialization. Usually it should be
    enough to have a single instance per program.
    """

    def __init__(self):
        """Initializes and loads the dictionary."""
        ##printmsg('Loading the dictionary.')
        self._dict = {}
        self._mixed_dict = {}

        with TextIOWrapper(resource_stream('lytspel', 'lytspel-dict.csv'),
                           encoding='utf-8') as csv_stream:
            csvreader = csv.reader(csv_stream)
            next(csvreader)  # skip header line
            redirects = {}  # will be resolved later

            for row in csvreader:
                tradspell = get_elem(row, 0)
                pos = get_elem(row, 1)
                redirect = get_elem(row, 2)
                lytspel = get_elem(row, 3)

                if tradspell and lytspel:
                    ts_lower = tradspell.lower()
                    ls_lower = lytspel.lower()

                    if pos:
                        # Treat value as a dict of POS-tagged entries
                        if not ts_lower in self._dict:
                            self._dict[ts_lower] = {}
                        self._dict[ts_lower][pos] = ls_lower
                    else:
                        self._dict[ts_lower] = ls_lower

                    if self.is_mixed_case(lytspel):
                        self._mixed_dict[tradspell] = lytspel

                elif tradspell and redirect:
                    redirects[tradspell.lower()] = redirect.lower()
                else:
                    warn('Unexpected/malformed CSV row: {}'.format(','.join(row)))

        # Resolve redirects
        for key, target in redirects.items():
            value = self._dict.get(target)
            if value:
                self._dict[key] = value
            else:
                warn("Target '{}' of redirect '{}' missing!".format(target, key))

    @staticmethod
    def is_mixed_case(word: str) -> bool:
        """Test whether a word is MiXed case.

        A word is assumed to be MiXed case if it contains at least one upper-case letter
        that's not the first letter and at least one lower-case letter. Words must be
        ASCII-fied for this function to work correctly.
        """
        return bool(CAPTALIZED_RE.search(word)) and bool(INNER_CAP_RE.search(word))

    @lru_cache(maxsize=1048576)
    def lookup(self, word: str, spacy_pos: str = None, at_sent_start: bool = False) -> Union[
            str, ConvState]:
        """Lookup a word in the dictionary and returned its converted form.

        'spacy_pos' is the POS tag as returned by spaCy. If it's omitted and if the respelling
        of the word depends on its POS tag, ConvState.NLP_NEEDED is returned.

        'at_sent_start' should be set to true if this token is the first of a new sentence.
        It is used to handle the case of the converted pronoun 'I' and its contractions.

        None is returned if the word is unknown.

        Case is restored (lower, Capitalized, or ALL_CAPS). MixedCase is also restored
        provided that *both* the input word and the dictionary entry use this case form
        (e.g. JavaScript -> JaavaScript).
        """
        # pylint: disable=too-many-branches
        word = self.asciify(word)
        # Strip final "'s" (genitive) or contraction and remember for later (handling case)
        word, contraction = self.strip_generic_contraction(word)

        if word in CAP_ABBREVS:
            result = word
        else:
            lower = word.lower()

            if lower in ("'ll", "'re", "'ve"):
                # Contraction tokens produced by spaCy: discard the last letter
                result = lower[:-1]
            elif SINGLE_LETTER_CONTRACTION_RE.match(lower):
                # Single-letter contraction token: return as is
                result = lower
            else:
                result = self._dict.get(lower)

                if result is None:
                    result = self.fallback_lookup(lower)
                    if result is None:
                        # Still no match
                        return None

                if isinstance(result, dict):
                    if spacy_pos:
                        result = self._find_pos_tagged_entry(result, spacy_pos)

                    else:
                        return ConvState.NLP_NEEDED

            result = self._restore_capitalization(result, word)

            if not at_sent_start and I_AND_CONTRACTIONS_RE.match(word) and result != 'I':
                # Correct the case of the pronoun 'I' and its contractions
                result = result.lower()

        if contraction:
            if len(contraction) == 3:
                # Strip the final letter of two-letter contractions such as "'ll" or "'ve"
                contraction = contraction[:-1]
            result += contraction

        result = result.replace("'", '’')  # replace normal by typographic apostrophe
        return result

    @staticmethod
    @lru_cache(maxsize=1048576)
    def asciify(word: str) -> str:
        """Convert a word to its ASCII equivalent.

        Typographic apostrophes are replaced by normal ones and diacritical letters are
        replaced by their closest ASCII equivalents.
        """
        # pylint: disable=too-many-branches, too-many-statements
        result = word.replace('’', "'")  # ASCII-ify apostrophes

        if NON_ASCII_LETTERS_RE.search(result): # non-ASCII word
            letters = list(result)

            for i, letter in enumerate(letters):
                if ord(letter) < 128:
                    continue  # ASCII, nothing to do

                lower_letter = letter.lower()
                replacement = ''

                # vowels and semivowels
                if lower_letter in 'áàăâǎåäãąāả':
                    replacement = 'a'
                elif lower_letter in 'æ':
                    replacement = 'ae'
                elif lower_letter in 'éèĕêěëẽęēẻ':
                    replacement = 'e'
                elif lower_letter in 'íìĭîǐïĩįī':
                    replacement = 'i'
                elif lower_letter in 'óòŏôǒöõøǫōỏ':
                    replacement = 'o'
                elif lower_letter in 'œ':
                    replacement = 'oe'
                elif lower_letter in 'úùŭûǔůüũųūủ':
                    replacement = 'u'
                elif lower_letter in 'ýỳŷẙÿỹȳỷ':
                    replacement = 'y'
                # consonants
                elif lower_letter in 'ćĉčċç':
                    replacement = 'c'
                elif lower_letter in 'ďđ':
                    replacement = 'd'
                elif lower_letter in 'ǵğĝǧġḡ':
                    replacement = 'g'
                elif lower_letter in 'ĥȟḧħ':
                    replacement = 'h'
                elif lower_letter in 'ĵǰ':
                    replacement = 'j'
                elif lower_letter in 'ḱǩ':
                    replacement = 'k'
                elif lower_letter in 'ĺľłḹ':
                    replacement = 'l'
                elif lower_letter in 'ḿ':
                    replacement = 'm'
                elif lower_letter in 'ńǹňñ':
                    replacement = 'n'
                elif lower_letter in 'ṕ':
                    replacement = 'p'
                elif lower_letter in 'ŕř':
                    replacement = 'r'
                elif lower_letter in 'śŝšş':
                    replacement = 's'
                elif lower_letter in 'ß':
                    replacement = 'ss'
                elif lower_letter in 'ťẗ':
                    replacement = 't'
                elif lower_letter in 'ṽ':
                    replacement = 'v'
                elif lower_letter in 'ẃẁẘẅ':
                    replacement = 'w'
                elif lower_letter in 'ẍ':
                    replacement = 'x'
                elif lower_letter in 'źẑž':
                    replacement = 'z'

                if replacement:
                    if letter != lower_letter:  #  original was upper case
                        replacement = replacement.upper()
                    letters[i] = replacement

            result = ''.join(letters)
        return result

    @staticmethod
    def strip_generic_contraction(word: str) -> Tuple[str, str]:
        """Strip the genitive ("'s") and generic contractions (such as "'ve") from a word.

        Returns two values: actual word, contraction. If the word doesn't end in a genitive
        or contraction, the first return value will be identical to the argument and the
        second return value will be empty.

        Both simple and typographic apostrophes are recognized.
        """
        match = GENERIC_CONTRACTION_RE.search(word)
        if match:
            idx = match.start() + 1
            return word[:idx], word[idx:]
        else:
            return word, ''

    def fallback_lookup(self, lc_word: str) -> str:
        """Fallback lookup if the regular lookup failed.

        'lc_word' must be a lower-case word.

        Checks if the word is a contraction that starts with an apostrophe (e.g. "'cause" or
        "'em"), as sometimes produced by spaCy, or if it ends in -in and a corresponding -ing
        entry exists (e.g. 'friggin' instead of 'frigging').
        """
        if lc_word.startswith("'"):
            result = self._dict.get(lc_word[1:])
            if isinstance(result, str):
                return "'" + result
        elif lc_word.endswith('in'):
            result = self._dict.get(lc_word + 'g')
            if isinstance(result, str) and result.endswith('g'):
                return result[:-1]

        return None

    @staticmethod
    def translate_pos(spacy_pos: str) -> str:
        """Translate a POS tag as used by spaCy into the form used in the dictionary."""
        if spacy_pos in ('AUX', 'VERB'):
            return 'v'
        elif spacy_pos in ('NOUN', 'PROPN'):
            return 'n'
        elif spacy_pos == 'ADJ':
            return 'aj'
        elif spacy_pos == 'ADV':
            return 'av'
        elif spacy_pos in ('ADP', 'PRON'):
            return 'prp'
        else:
            # Other POS tags shouldn't usually occur regarding our POS-tagged words,
            # but we use a reasonable default for cases where spaCy gets it wrong
            return 'v'

    def _find_pos_tagged_entry(self, entries: dict, pos: str,
                               tried_variants: List[str] = None) -> str:
        """Find the entry to use if several POS-tagged spellings exist for a word.

        'pos' is the POS tag as returned by spaCy (capitalized) or used in our dictionary
        (lower-case) - the former will be translated into the later if needed.

        'tried_variants' is used internally to prevent endless loops in case another POS tag
        has to be tried as fallback.
        """
        if pos and pos[0].isupper():
            pos = self.translate_pos(pos)
        if tried_variants is None:
            tried_variants = []

        tried_variants.append(pos)
        result = entries.get(pos)

        if result is None:
            fallback = POS_FALLBACKS.get(pos)

            if fallback is None or fallback in tried_variants:
                if fallback is None:
                    warn("No fallback found for POS tag '{}'; using last entry of {} as fallback"
                         .format(pos, entries))
                else:
                    warn("Trying POS tag '{}' after {} would loop; using last entry of {} as "
                         'fallback'.format(pos, tried_variants, entries))

                result = entries[sorted(entries.keys())[-1]]
            else:
                result = self._find_pos_tagged_entry(entries, fallback, tried_variants)

        return result

    def _restore_capitalization(self, converted: str, word: str) -> str:
        """Restore the capitalization of 'converted' to match that of 'word'."""
        if ALL_CAPS_RE.match(word):
            return converted.upper()  # ALL_CAPS

        if word in self._mixed_dict:
            return self._mixed_dict[word]  #  MixedCase

        if UPPER_CASE_RE.match(word):
            return converted[0].upper() + converted[1:]  # Capitalized

        return converted  # No change



"""Utility functions for Lytspel."""

from sys import stderr
from typing import Sequence, TypeVar


PACKAGENAME = 'lytspel'

T = TypeVar('T')


def get_elem(seq: Sequence[T], idx: int) -> T:
    """Safely retried an element from a sequence.

    None is returned if 'seq' ends before the requested 'idx' position.
    """
    if len(seq) > idx:
        return seq[idx]
    else:
        return None


def printmsg(msg: str) -> None:
    """Print a message to stderr, prefixed by the name of the script."""
    print('{}: {}'.format(PACKAGENAME, msg), file=stderr)


def readfile(filename: str) -> str:
    """Read a whole file into a string and return it."""
    with open(filename) as file:
        return file.read()



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='lytspel',
    version='1.1.0',
    author='Christian Siefkes',
    author_email='christian@siefkes.net',
    description='A Simple Phonetic Respelling for the English Language',
    entry_points = {
        'console_scripts': ['lytspel=lytspel:main'],
    },
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='ISC',
    url='https://www.lytspel.org/',
    packages=setuptools.find_packages(),
    install_requires=[
        'lxml >= 4.0.0',
        'setuptools >= 34.3.3',
        'spacy >= 2.0.0',
    ],
    # Keep minimum version listed here in sync with the one in INSTALL-USE.md
    python_requires='>=3.5',
    package_data={'lytspel': ['lytspel-dict.csv']},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Communications',
        'Topic :: Text Processing :: Linguistic',
    ],
    keywords='english spelling reform lytspel',
    project_urls={
        'Source': 'https://github.com/ChristianSi/lytspel',
        'Tracker': 'https://github.com/ChristianSi/lytspel/issues',
    },
 )
